/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2005 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "StdAfx.h"
#include "DesktopItem.h"
#include "DesktopTexture.h"
#include "Font.h"
#include "ContextMenu.h"

#define TIMER_MOUSETRACK 1

extern DesktopItemList desktopitemList;

extern Font *defaultfont;
extern DesktopTexture *defaultbackground;

extern int screenX;
extern int screenY;

extern int bitsperpixel;

extern bool disableSystemContext;

DesktopItem::DesktopItem(const string &name):actualheight(0),actualwidth(0),actualx(0),actualy(0)
{
	this->name = name;

	hWnd = 0;
	hInstance = 0;

	TooltipHints = 0;

	backupDC = 0;
	backupBitmap = 0;
	oldbackupBMP = 0;
	
	mousePressed = false;
	mouseInside = false;

	iconOverlay = false;

	DIsnapto = false;

	visible = false;
}

DesktopItem::~DesktopItem()
{
	if (hWnd && TooltipHints)
	{
		removeHint(hWnd);
		DestroyWindow(TooltipHints);
		TooltipHints = 0;
	}

	if(IsWindow(hWnd))
	{
		hide();

		KillTimer(hWnd, TIMER_MOUSETRACK);

		DestroyWindow(hWnd);
	}

	if(backupBitmap)
	{
		backupBitmap = (HBITMAP) SelectObject(backupDC, oldbackupBMP);
		DeleteObject(backupBitmap);
	}
	if (backupDC)
		DeleteDC(backupDC);
}

void DesktopItem::desktopload(HINSTANCE hInstance, string configLine)
{
	this->hInstance = hInstance;
	
	if (IsWindow(hWnd))
		return;

	hWnd = CreateWindowEx(WS_EX_TOOLWINDOW,
		"xDesktopItem",
		name.c_str(),
		WS_POPUP,
		actualx, actualy,			
		actualwidth, actualheight,
		GetLitestepWnd(),  
		0,
		hInstance,
		this);

	if (hWnd)
	{
		SetWindowLong(hWnd, GWL_USERDATA, magicDWord);
		
		bool balloontooltip = false;
		if ( IsOS(OS_2KXP) && GetxRCBool("xDesktop", "BalloonTooltip") )
			balloontooltip = true;

		// struct specifying control classes to register
		INITCOMMONCONTROLSEX iccex; 

		/* INITIALIZE COMMON CONTROLS */
		iccex.dwICC = ICC_WIN95_CLASSES;
		iccex.dwSize = sizeof(INITCOMMONCONTROLSEX);

		InitCommonControlsEx(&iccex);
		
		//ToolTips for Icons
		TooltipHints = CreateWindowEx(WS_EX_TOOLWINDOW, TOOLTIPS_CLASS, TEXT(""), WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP | (balloontooltip ? TTS_BALLOON : 0),
									CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
									NULL, NULL, hInstance, NULL);

		if (TooltipHints)
		{
			SetWindowPos(TooltipHints, HWND_TOPMOST, 0, 0, 0, 0,
						 SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

			SendMessage(TooltipHints, TTM_SETMAXTIPWIDTH, 0, 600);

			SendMessage( TooltipHints, (UINT) TTM_SETTIPBKCOLOR, (WPARAM) (COLORREF) GetxRCColor("xDesktop", "ToolTipColor", GetSysColor(COLOR_INFOBK)), 0 );   
			SendMessage( TooltipHints, (UINT) TTM_SETTIPTEXTCOLOR, (WPARAM) (COLORREF) GetxRCColor("xDesktop", "ToolTipTextColor", GetSysColor(COLOR_INFOTEXT)), 0);  

			char cone[32] = {0};
			char ctwo[32] = {0};
			char *cbuffers[] = {cone, ctwo};
			
			int numTokens = LCTokenize( (GetxRCLine("xDesktop", "TooltipDurations", "-1 -1")).c_str(), cbuffers, 2, NULL);
		
			int initial = -1;
			int autopop = -1;

			if ( numTokens >= 1 )
			{
				initial = atoi(cone);
				if ( numTokens >= 2 )
					autopop = atoi(ctwo);
			}

			if ( initial <= 0 )
				initial = -1;
			if ( autopop <= 1000 )
				autopop = -1;

			if ( initial != -1 )
				SendMessage( TooltipHints, (UINT) TTM_SETDELAYTIME,	(WPARAM) (DWORD) TTDT_INITIAL, (LPARAM) MAKELONG(initial, 0) );
			if ( autopop != -1 )
				SendMessage( TooltipHints, (UINT) TTM_SETDELAYTIME,	(WPARAM) (DWORD) TTDT_AUTOPOP, (LPARAM) MAKELONG(autopop, 0) );
		
			if ( balloontooltip )
				SendMessage( TooltipHints, (UINT) TTM_SETTITLE, (WPARAM) (int) 1, (LPARAM) "Icon Info:" );
		}

		desktopreconfigure(configLine);
	}
	else
	{
		delete this;
	}
}

void DesktopItem::desktopreconfigure(string configLine)
{
		SetWindowRgn(hWnd, NULL, true);
	
		NameValuePair iconTextValues[] = {
			{ "top", 1 },
			{ "bottom", 2 },
			{ "left", 3 },
			{ "right", 4 },
			{ 0, 0 }
		};

		//Initialize Helper Strings
		int numTokens = 0;
		string one = "";
		string two = "";
		string three = "";
		string four = "";
		string five = "";

		char path[MAX_PATH_LENGTH] = {0};

		//Painting Background
		alphaTransparency = GetxRCInt("xDesktop" , "AlphaTransparency", 255, 0, 255);
		activealphaTransparency = GetxRCInt("xDesktop" , "ActiveAlphaTransparency", alphaTransparency, 0, 255);
		//Prepare Settings
		//----------------------------------------------------------------
		SplitString(configLine, one, two, three, four, five);

		actualx = atoi( two.c_str() );
		actualy = atoi( three.c_str() );
		actualwidth = GetxRCInt("xDesktop", "ItemWidth", 100);
		actualheight = GetxRCInt("xDesktop", "ItemHeight", 75);
		iconCommand = four;
		strcpy(path, four.c_str());
		positionSaver = (IntToString(actualx) + " " + IntToString(actualy) + " \"" + four + "\"");	
		//----------------------------------------------------------------

		//SnapTo Settings
		numTokens = SplitString(GetxRCLine("xDesktop", "SnapTo", ""), one, two, three, four, five);

		if ( numTokens == 5 )
		{
			DIsnapto = true;
			DIx = atoi( one.c_str() );
			DIy = atoi( two.c_str() );
			DIwidth = atoi( three.c_str() );
			DIheight = atoi( four.c_str() );
			DIsnapdistance = atoi( five.c_str() );
		}

		//Prepare Special DesktopSettings
		//----------------------------------------------------------------
		iconTextSpacing = GetxRCInt("xDesktop", "TextSpacing", 5, 0);
		iconSize = GetxRCInt("xDesktop", "IconSize", 32, 8, 64);
		iconTextAlign = GetxRCNamedValue("xDesktop", "TextPosition", iconTextValues, 2);
		textBorder = GetxRCInt("xDesktop", "TextBorder", 3, 0);
		noWordBreak = GetxRCBool("xDesktop", "NoLineBreak");
		//----------------------------------------------------------------

		bool itemIsFolder = false;
		
		HRESULT hres;
		IShellLink* psl;
		SHFILEINFO shFileInfo;

		char szGetPath[MAX_PATH_LENGTH];
		if (strstr(iconCommand.c_str(), "::{") != NULL)
		{
			itemIsFolder = true;
			SHGetFileInfo(iconCommand.c_str(), NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_DISPLAYNAME | SHGFI_TYPENAME);
			text = trim(shFileInfo.szDisplayName);
			strcpy(tooltip , "Type:    ");
			strcat(tooltip , shFileInfo.szTypeName);
			strcat(tooltip, "\nTarget: ");
			strcat(tooltip , shFileInfo.szDisplayName);
		}
		else
		{
			SHGetFileInfo(iconCommand.c_str(), NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_DISPLAYNAME | SHGFI_TYPENAME | SHGFI_ATTRIBUTES);
			text = trim(shFileInfo.szDisplayName);
			strcpy(tooltip , "Type:    ");
			strcat(tooltip , shFileInfo.szTypeName);
		
			if (shFileInfo.dwAttributes & SFGAO_FOLDER)
				itemIsFolder = true;
		}

		//Create the ShellLink object
		hres = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER, IID_IShellLink, (LPVOID*) &psl);

		if (SUCCEEDED(hres))
		{
			IPersistFile* ppf;
			//Bind the ShellLink object to the Persistent File
			hres = psl->QueryInterface( IID_IPersistFile, (LPVOID *) &ppf);

			if (SUCCEEDED(hres))
			{
				WORD wsz[MAX_PATH_LENGTH];
				//Get a UNICODE wide string wsz from the Link path
				MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, path, -1, wsz, MAX_PATH_LENGTH);
					
				//Read the link into the persistent file
				hres = ppf->Load(wsz, 0);
					
				if (SUCCEEDED(hres))
				{
					//Read the target information from the link object
					//UNC paths are supported (SLGP_UNCPRIORITY)
					psl->GetPath(szGetPath, MAX_PATH_LENGTH, NULL, SLGP_UNCPRIORITY);
						
					strcat(tooltip, "\nTarget: ");
					strcat(tooltip, szGetPath);

					psl->GetDescription(szGetPath, MAX_PATH_LENGTH);

					if (stricmp(szGetPath, "") != NULL)
					{
						strcat(tooltip, "\n\nDescription:\n");
						strcat(tooltip, szGetPath);

						SendMessage(TooltipHints, TTM_SETMAXTIPWIDTH, 0, 300);
					}
				}
				else if (strstr(path, "::{") == NULL)
				{
					strcat(tooltip, "\nName:   ");
					strcat(tooltip, iconCommand.c_str());
				}
			}
		}
		psl->Release();

		addHint(hWnd, tooltip);

		//Events/Actions
		easyDesktopFocus = GetxRCBool("xDesktop", "EasyFocus");
		
		leftClickCommand = GetxRCString("xDesktop", "OnLeftClick", ".move");
		leftDoubleClickCommand  = GetxRCString("xDesktop", "OnLeftDoubleClick", ".open");
		middleClickCommand = GetxRCString("xDesktop", "OnMiddleClick", ".properties");
		middleDoubleClickCommand = GetxRCString("xDesktop", "OnMiddleDoubleClick", ".none");
		rightClickCommand =  GetxRCString("xDesktop", "OnRightClick", ".context");
		rightDoubleClickCommand =  GetxRCString("xDesktop", "OnRightDoubleClick", ".none");

		if (itemIsFolder)
		{
			if (strstr(iconCommand.c_str(), "::{20D04FE0-3AEA-1069-A2D8-08002B30309D}") != NULL)
			{
				dropCommand = "";
				itemIsFolder = false;
			}
			else if (strstr(iconCommand.c_str(), "::{450D8FBA-AD25-11D0-98A8-0800361B1103}") != NULL)
			{
				SHGetSpecialFolderPath(NULL, path, CSIDL_PERSONAL, false);
				dropCommand = "copy \"";
				dropCommand.append( path );
				dropCommand.append( "\"" );
			}
			else if (strstr(iconCommand.c_str(), "::{208D2C60-3AEA-1069-A2D7-08002B30309D}") != NULL)
			{
				dropCommand = "";
				itemIsFolder = false;
			}
			else if (strstr(iconCommand.c_str(), "::{645FF040-5081-101B-9F08-00AA002F954E}") != NULL)
			{
				dropCommand = "delete files";
			}
			else
				dropCommand = ("copy \"" + iconCommand + "\"");
		}
		else
		{
			dropCommand = "";
		}

		//Moving
		moveable = GetxRCBool("xDesktop", "Moveable", true);
		//Apply Selection of MoveButton
		if (stricmp(middleClickCommand.c_str(), ".move") == NULL)
			moveButton = 3;

		if (stricmp(rightClickCommand.c_str(), ".move") == NULL)
			moveButton = 2;

		if (stricmp(leftClickCommand.c_str(), ".move") == NULL)
			moveButton = 1;

		if (moveButton == 0)
			moveable = false;

		//Apply Selection of clicks
		if (stricmp(leftClickCommand.c_str(), ".open") == NULL && !itemIsFolder)
			leftClickCommand = ("!execute [\"" + iconCommand + "\"]");
		else if (stricmp(leftClickCommand.c_str(), ".open") == NULL && itemIsFolder)
			leftClickCommand = ("!execute [explorer.exe /e,\"" + iconCommand + "\"]");
		else if (stricmp(leftClickCommand.c_str(), ".context") != NULL && stricmp(leftClickCommand.c_str(), ".properties") != NULL)
			leftClickCommand = "";

		if (stricmp(leftDoubleClickCommand.c_str(), ".open") == NULL && !itemIsFolder)
			leftDoubleClickCommand = ("!execute [\"" + iconCommand + "\"]");
		else if (stricmp(leftDoubleClickCommand.c_str(), ".open") == NULL && itemIsFolder)
			leftDoubleClickCommand = ("!execute [explorer.exe /e,\"" + iconCommand + "\"]");
		else
			leftDoubleClickCommand = "";

		if (stricmp(middleClickCommand.c_str(), ".open") == NULL && !itemIsFolder)
			middleClickCommand = ("!execute [\"" + iconCommand + "\"]");
		else if (stricmp(middleClickCommand.c_str(), ".open") == NULL && itemIsFolder)
			middleClickCommand = ("!execute [explorer.exe /e,\"" + iconCommand + "\"]");
		else if (stricmp(middleClickCommand.c_str(), ".context") != NULL && stricmp(middleClickCommand.c_str(), ".properties") != NULL)
			middleClickCommand = "";
		
		if (stricmp(middleDoubleClickCommand.c_str(), ".open") == NULL && !itemIsFolder)
			middleDoubleClickCommand = ("!execute[\"" + iconCommand + "\"]");
		else if (stricmp(middleDoubleClickCommand.c_str(), ".open") == NULL && itemIsFolder)
			middleDoubleClickCommand = ("!execute [explorer.exe /e,\"" + iconCommand + "\"]");
		else
			middleDoubleClickCommand = "";

		if (stricmp(rightClickCommand.c_str(), ".open") == NULL && !itemIsFolder)
			rightClickCommand = ("!execute [\"" + iconCommand + "\"]");
		else if (stricmp(rightClickCommand.c_str(), ".open") == NULL && itemIsFolder)
			rightClickCommand = ("!execute [explorer.exe /e,\"" + iconCommand + "\"]");
		else if (stricmp(rightClickCommand.c_str(), ".context") != NULL && stricmp(rightClickCommand.c_str(), ".properties") != NULL)
			rightClickCommand = "";
		
		if (stricmp(rightDoubleClickCommand.c_str(), ".open") == NULL && !itemIsFolder)
			rightDoubleClickCommand = ("!execute [\"" + iconCommand + "\"]");
		else if (stricmp(rightDoubleClickCommand.c_str(), ".open") == NULL && itemIsFolder)
			rightDoubleClickCommand = ("!execute [explorer.exe /e,\"" + iconCommand + "\"]");
		else
			rightDoubleClickCommand = "";

		//Initial Repositioning
		reposition(actualx, actualy , actualwidth, actualheight);
		
		//Tooltip
		addHint(hWnd, tooltip);

		//File Actions
		if (itemIsFolder)
			DragAcceptFiles(hWnd, true);
		else
			DragAcceptFiles(hWnd, false);

		ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
		SetParent(hWnd, GetLitestepDesktop());
		ModifyStyle(hWnd, WS_CHILD, WS_POPUP);
			
		SetWindowPos(hWnd, HWND_BOTTOM,
				0, 0, 0, 0,
				SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
			
		//Initial Show
		GetxRCBool("xDesktop", "StartHidden") ? hide() : show();
}

//------------------------------------------------------------
//Reposition
//------------------------------------------------------------

void DesktopItem::reposition(int newx, int newy, int newwidth, int newheight)
{
	if (newheight == 0 && newwidth == 0)
	{
		SetWindowPos(hWnd, 0, newx, newy, 0, 0,
				SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
	}
	else
	{
		SetWindowPos(hWnd, 0, newx, newy, newwidth, newheight,
				SWP_NOACTIVATE | SWP_NOZORDER);
	}
}

//------------------------------------------------------------
//Show/Hide
//------------------------------------------------------------
void DesktopItem::show()
{
	ShowWindow(hWnd, SW_SHOWNOACTIVATE);

	visible = true;

	repaint(true);
}

void DesktopItem::hide()
{
	ShowWindow(hWnd, SW_HIDE);
		
	visible = false;
}

//------------------------------------------------------------
//Mouse Related Actions
//------------------------------------------------------------

void DesktopItem::onLButtonDblClk(int x, int y)
{
	if(!leftDoubleClickCommand.empty())
		LSExecute(hWnd, leftDoubleClickCommand.c_str(), SW_SHOWNORMAL);
}

void DesktopItem::onLButtonDown(int x, int y)
{
	// 1 = LeftButton
	ontemplateButtonDown(1);
}

void DesktopItem::onLButtonUp(int x, int y)
{	
	// 1 = LeftButton
	ontemplateButtonUp(x, y, 1, leftClickCommand);
}

void DesktopItem::onMButtonDblClk(int x, int y)
{
	if(!middleDoubleClickCommand.empty())
		LSExecute(hWnd, middleDoubleClickCommand.c_str(), SW_SHOWNORMAL);
}

void DesktopItem::onMButtonDown(int x, int y)
{
	// 3 = MiddleButton
	ontemplateButtonDown(3);
}

void DesktopItem::onMButtonUp(int x, int y)
{
	// 3 = MiddleButton
	ontemplateButtonUp(x, y, 3, middleClickCommand);
}

void DesktopItem::onRButtonDblClk(int x, int y)
{
	if(!rightDoubleClickCommand.empty())
		LSExecute(hWnd, rightDoubleClickCommand.c_str(), SW_SHOWNORMAL);
}

void DesktopItem::onRButtonDown(int x, int y)
{
	// 2 = RightButton
	ontemplateButtonDown(2);
}

void DesktopItem::onRButtonUp(int x, int y)
{
	// 2 = RightButton
	ontemplateButtonUp(x, y, 2, rightClickCommand);
}

void DesktopItem::ontemplateButtonDown(int actualMouseClick)
{
	if(moveable && moveButton == actualMouseClick)
	{
		if (!mousePressed)
		{
			mousePressed = true;
			GetCursorPos(&savedpt);
			ScreenToClient(hWnd, &savedpt);
			SetCapture(hWnd);
		}
	}
}

void DesktopItem::ontemplateButtonUp(const int &x, const int &y, int actualMouseClick, const string &clickCommand)
{
	if (moveable && moveButton == actualMouseClick)
	{	
		//To be on the save side
		ReleaseCapture();

		mousePressed = false;

		if ( alphaTransparency != 255 )
			repaint(true, true);
	}
	else 
	{
		//To be on the save side
		ReleaseCapture();

		if (!clickCommand.empty() && stricmp(clickCommand.c_str(), ".context") != NULL && stricmp(clickCommand.c_str(), ".properties") != NULL)
		{
			LSExecute(hWnd, clickCommand.c_str(), SW_SHOWNORMAL);
		}
		else if (stricmp(clickCommand.c_str(), ".context") == NULL)
		{
			if ( !disableSystemContext || (disableSystemContext && strstr(iconCommand.c_str() , "::{") == NULL) )
				ContextMenu( iconCommand.c_str() );
		}
		else if (stricmp(clickCommand.c_str(), ".properties") == NULL)
		{
			SHELLEXECUTEINFO ShExecInfo = {0};
			ShExecInfo.cbSize = sizeof(SHELLEXECUTEINFO);
			ShExecInfo.fMask = SEE_MASK_INVOKEIDLIST;
			ShExecInfo.hwnd = NULL;
			ShExecInfo.lpVerb = "properties";
			ShExecInfo.lpFile = iconCommand.c_str();
			ShExecInfo.lpParameters = ""; 
			ShExecInfo.lpDirectory = NULL;
			ShExecInfo.nShow = SW_SHOW;
			ShExecInfo.hInstApp = NULL; 
			ShellExecuteEx(&ShExecInfo);
		}

		mousePressed = false;
	}
}

void DesktopItem::onMouseMove(int x, int y)
{
	if (!mouseInside)
	{
		mouseInside = true;

		if (easyDesktopFocus)
			SendMessage(hWnd, WM_ACTIVATE, WA_CLICKACTIVE, 0);

		SetTimer(hWnd, TIMER_MOUSETRACK, 100, 0);
	}

	if (moveable && mousePressed)
	{
		KillTimer(hWnd, TIMER_MOUSETRACK);

		POINT newpt = {x,y};
		ClientToScreen(hWnd, &newpt);
		newpt.x = newpt.x-savedpt.x;
		newpt.y = newpt.y-savedpt.y;

		reposition(newpt.x, newpt.y, 0, 0);
	}
	else
	{
		SetTimer(hWnd, TIMER_MOUSETRACK, 100, 0);
	}
}

void DesktopItem::repaint(bool invalidateCache, bool xpMove)
{
	if ( actualwidth <= 0 || actualheight <= 0 )
		return;

	if (invalidateCache)
	{
		if ( backupBitmap )
		{
			backupBitmap = (HBITMAP) SelectObject(backupDC, oldbackupBMP);
			DeleteObject(backupBitmap);
		}
		if ( backupDC )
			DeleteObject(backupDC);
				
		backupBitmap = 0;
		backupDC = 0;
	}
	
	RedrawWindow(hWnd, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW);

	//Mostly except Move under XP ( repaint needed due to BG repainting, but no Region Change )
	if ( !xpMove )
		updateRegion();	
}

void DesktopItem::onPaint(HDC hDC)
{
	if ( actualwidth <= 0 || actualheight <= 0 )
		return;

	int width = actualwidth;
	int height = actualheight;

	HDC backgroundDC = 0;
	HBITMAP backgroundBitmap = 0;
	HBITMAP oldBMP = 0;

	// Use Complete Backup for WindowsInternal Repaint (Covering Windows, ...)
	if ( backupBitmap != 0 && backupDC != 0 )
	{
		BitBlt(hDC, 0, 0, width, height, backupDC, 0, 0, SRCCOPY);
	}
	else
	//1----------------------------------------------------------------
	{
	//1----------------------------------------------------------------

		RECT mainRect;
		GetWindowRect(hWnd, &mainRect);

		width = mainRect.right - mainRect.left;
		height = mainRect.bottom - mainRect.top;

		backgroundDC = CreateCompatibleDC(hDC);
		backgroundBitmap = CreateCompatibleBitmap(hDC, width, height);
		oldBMP = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
		
		POINT maxpaintsize;

		if ( iconTextAlign == 1 || iconTextAlign == 2 )
		{
			maxpaintsize.x = width-(2*textBorder);
			maxpaintsize.y = height-(2*textBorder)-iconSize-iconTextSpacing;
		}
		else if ( iconTextAlign == 3 || iconTextAlign == 4 )
		{	
			maxpaintsize.x = width-(2*textBorder)-iconSize-iconTextSpacing;
			maxpaintsize.y = height-(2*textBorder);
		}

		//Normal Background Painting
		if ( defaultbackground )
		{
			if ( defaultfont )
			{
				maxpaintsize = defaultfont->measure(backgroundDC,
												0,
												0,
												maxpaintsize.x,
												maxpaintsize.y,
												text,
												DT_CENTER | DT_NOPREFIX | (noWordBreak ? DT_SINGLELINE : DT_WORDBREAK ),
												iconOverlay);
			}

			maxpaintsize.x = maxpaintsize.x+textBorder*2;
			maxpaintsize.y = maxpaintsize.y+textBorder*2;

			defaultbackground->apply(backgroundDC, 0, 0, width, height, maxpaintsize, iconCommand, name, iconOverlay);
		}

		if ( iconTextAlign == 1 || iconTextAlign == 2 )
		{
			maxpaintsize.x = width-(2*textBorder);
			maxpaintsize.y = min(maxpaintsize.y-textBorder*2, actualheight-(2*textBorder));
		}
		else if ( iconTextAlign == 3 || iconTextAlign == 4 )
		{
			maxpaintsize.x = maxpaintsize.x;
			maxpaintsize.y = maxpaintsize.y-(2*textBorder);
		}
	
		//--------------------------------------------------------------------------------
		// start rendering text
		//--------------------------------------------------------------------------------

		if ( defaultfont )
		{
			if ( iconTextAlign == 1 )
			{
				defaultfont->apply(backgroundDC,
											(width-maxpaintsize.x)/2,
											max(height-iconSize-iconTextSpacing-textBorder-maxpaintsize.y, 0),
											maxpaintsize.x,
											maxpaintsize.y,
											text,
											DT_CENTER | DT_NOPREFIX | (noWordBreak ? DT_SINGLELINE : DT_WORDBREAK),
											iconOverlay);
			}
			else if ( iconTextAlign == 2 )
			{
				defaultfont->apply(backgroundDC,
										(width-maxpaintsize.x)/2,
										iconSize+iconTextSpacing+textBorder,
										maxpaintsize.x,
										maxpaintsize.y,
										text,
										DT_CENTER | DT_NOPREFIX | (noWordBreak ? DT_SINGLELINE : DT_WORDBREAK),
										iconOverlay);
			}
			else if ( iconTextAlign == 3 )
			{
				defaultfont->apply(backgroundDC,
										max(width-iconSize-iconTextSpacing-maxpaintsize.x+textBorder, textBorder),
										(height-maxpaintsize.y)/2,
										min(maxpaintsize.x-2*textBorder, width-iconSize-iconTextSpacing-(2*textBorder)),
										maxpaintsize.y,
										text,
										DT_CENTER | DT_NOPREFIX | (noWordBreak ? DT_SINGLELINE : DT_WORDBREAK),
										iconOverlay);
			}
			else if ( iconTextAlign == 4 )
			{
				defaultfont->apply(backgroundDC,
										iconSize+iconTextSpacing+textBorder,
										(height-maxpaintsize.y)/2,
										min(maxpaintsize.x-2*textBorder, width-iconSize-iconTextSpacing-(2*textBorder)),
										maxpaintsize.y,
										text,
										DT_CENTER | DT_NOPREFIX | (noWordBreak ? DT_SINGLELINE : DT_WORDBREAK),
										iconOverlay);
			}
		}

		//--------------------------------------------------------------------------------
		// stop rendering text
		//--------------------------------------------------------------------------------

		//--------------------------------------------------------------------------------
		// start fake alpha
		//--------------------------------------------------------------------------------
		
		if ( (!iconOverlay && alphaTransparency != 255) || (iconOverlay && activealphaTransparency != 255) )
		{
			HDC memoryDesktopDC = CreateCompatibleDC(hDC);
			HBITMAP memoryDesktopBitmap = CreateCompatibleBitmap(hDC, width, height);
			HBITMAP oldmemoryBMP = (HBITMAP) SelectObject(memoryDesktopDC, memoryDesktopBitmap);

			PaintDesktopEx(memoryDesktopDC, 0, 0, width, height, actualx, actualy, false);

			int usedAlpha = (iconOverlay ? activealphaTransparency : alphaTransparency);
			int iy = 0;
			int ix = 0;
			COLORREF cl = 0;
			COLORREF clD = 0;
			BYTE r = 0;
			BYTE g = 0;
			BYTE b = 0;
			BYTE rD = 0;
			BYTE gD = 0;
			BYTE bD = 0;
			for(iy = 0; iy < height; iy++)
			{
				for(ix = 0; ix < width; ix++)
				{
					cl = GetPixel(backgroundDC, ix, iy);
					// Loop through all transparent pixels...
					if( cl != RGB(255,0,255) )
					{
						r = GetRValue(cl);
						g = GetGValue(cl);
						b = GetBValue(cl);

						clD = GetPixel(memoryDesktopDC, ix, iy);

						rD = GetRValue(clD);
						gD = GetGValue(clD);
						bD = GetBValue(clD);

						// Fake Alpha Effect
						r = (BYTE)((rD * (255 - usedAlpha) + r * usedAlpha + 255) >> 8);
						g = (BYTE)((gD * (255 - usedAlpha) + g * usedAlpha + 255) >> 8);
						b = (BYTE)((bD * (255 - usedAlpha) + b * usedAlpha + 255) >> 8);

						SetPixelV(backgroundDC, ix, iy, RGB(r, g, b));
					}
				}
			}
			
			memoryDesktopBitmap = (HBITMAP) SelectObject(memoryDesktopDC, oldmemoryBMP);
			DeleteObject(memoryDesktopBitmap);
			DeleteDC(memoryDesktopDC);
		}

		//--------------------------------------------------------------------------------
		// stop fake alpha
		//--------------------------------------------------------------------------------

		// blt the double buffer to the display
		BitBlt(hDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);

		//Create BackupBitmap, if it doesn't exist!
		//----------------------------------------------------------------------
		if ( backupBitmap == 0 && backupDC == 0 )
		{
			backupDC = CreateCompatibleDC(hDC);
			backupBitmap = CreateCompatibleBitmap(hDC, width, height);
			oldbackupBMP = (HBITMAP) SelectObject(backupDC, backupBitmap);

			BitBlt(backupDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);
		}
		//----------------------------------------------------------------------

	//1----------------------------------------------------------------
	}
	//1----------------------------------------------------------------

	if ( backgroundBitmap != 0 && backgroundDC != 0 )
	{
		backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, oldBMP);
		DeleteObject(backgroundBitmap);
		DeleteDC(backgroundDC);
	}
}

void DesktopItem::updateRegion()
{
	HDC tmpDC = CreateCompatibleDC(NULL);
	HBITMAP tmpBmp = CreateBitmap(actualwidth, actualheight, 1, bitsperpixel, NULL);
	HBITMAP oldBMP = (HBITMAP) SelectObject(tmpDC, tmpBmp);

	onPaint(tmpDC);
								
	tmpBmp = (HBITMAP) SelectObject(tmpDC, oldBMP);

	HRGN region = BitmapToRegion(tmpBmp, RGB(255, 0, 255), 0, 0, 0);
	HRGN windowRgn = CreateRectRgn(0, 0, 0, 0);
	CombineRgn(windowRgn, region, NULL, RGN_COPY);

	SetWindowRgn(hWnd, windowRgn, true);
	
	DeleteObject(windowRgn);
	DeleteObject(region);

	DeleteObject(tmpBmp);
	DeleteDC(tmpDC);
}

void DesktopItem::onWindowPosChanging(WINDOWPOS *windowPos)
{
	if ( mousePressed && DIsnapto && (windowPos->flags & SWP_NOMOVE) == 0 )
	{
		POINT pos = {windowPos->x, windowPos->y};

		int result = 0;

		for( int i = 0; i != (screenX-2*DIx)/DIwidth; i++)
		{
			if ( pos.x > DIx+i*DIwidth+DIsnapdistance )
				continue;

			for (int j = 0; j != (screenY-2*DIy)/DIheight; j++)
			{
				if ( pos.y > DIy+j*DIheight+DIsnapdistance )
					continue;

				result = isCloseTo(pos, DIx+i*DIwidth, DIy+j*DIheight, DIsnapdistance);
						
				if (result != 0)
					break;
			}
			if (result != 0)
				break;
		}
				
		if (result == 1)
		{
			windowPos->x = pos.x;
		}
		else if (result == 2)
		{
			windowPos->y = pos.y;
		}
		else if (result == 3)
		{
			windowPos->x = pos.x;
			windowPos->y = pos.y;
		}
	}

	//Remove Zorder usage
	if ( (windowPos->flags & SWP_NOZORDER) == 0 )
		windowPos->flags = windowPos->flags | SWP_NOZORDER;
}

void DesktopItem::onMove(int x, int y)
{
	actualx = x;
	actualy = y;

	//Save New Position
	positionSaver = (IntToString( actualx ) + " " + IntToString( actualy ) + " \"" + iconCommand + "\"");

	if ( IsOS( OS_XP ) )
		repaint(true, true);
}

void DesktopItem::onSize(int width, int height)
{
	actualheight = height;
	actualwidth = width;

	repaint(true);
}

void DesktopItem::onTimer(int timerID)
{
	switch(timerID)
	{
		case TIMER_MOUSETRACK:
		{
			POINT pt;
			RECT rc;
			GetCursorPos(&pt);
			GetWindowRect(hWnd, &rc);

			if(!PtInRect(&rc, pt))
			{
				//To be on the save side
				ReleaseCapture();

				KillTimer(hWnd, TIMER_MOUSETRACK);
				mouseInside = false;
				mousePressed = false;

				if (easyDesktopFocus)
					SendMessage(hWnd, WM_ACTIVATE, WA_INACTIVE, 0);
			}
		}
		break;

		default:
		break;
	}
}

bool DesktopItem::onWindowMessage(UINT message, WPARAM wParam, LPARAM lParam, LRESULT &lResult)
{
	extern HWND messageHandler;

	switch(message)
	{
		case WM_ACTIVATE:
		{
			if (wParam == WA_INACTIVE)
			{
				iconOverlay = false;
				repaint(true);
			}
			else if (wParam == WA_CLICKACTIVE)
			{
				iconOverlay = true;
				repaint(true);
			}

			return true;
		}

		case WM_CLOSE:
		{
			lResult = 0;
			return true;
		}

		case WM_DROPFILES:
		{
			if(dropCommand.empty())	// we shouldn't get here... but somehow we did :p
				break;

			int intFileAction;

			char action[32], szTarget[MAX_PATH];;
			char *buffers[] = {action, szTarget};

			LCTokenize(dropCommand.c_str(), buffers, 2, NULL);

			if (stricmp(action, "move") == 0)
				intFileAction = 1;
			else if (stricmp(action, "copy") == 0)
				intFileAction = 2;
			else if (stricmp(action, "delete") == 0)
				intFileAction = 0;

			char *szSourceFiles = NULL;
			char buffer[MAX_PATH];
			int numFiles;
			UINT len,scursor = 0;

			numFiles = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, NULL);

			szSourceFiles = new char[MAX_PATH*numFiles+1];
			scursor = 0;

			for (int i=0;i<numFiles;i++)
			{
				len = DragQueryFile((HDROP) wParam, i, buffer, MAX_PATH);
					
				strcpy(szSourceFiles+scursor,buffer);
				scursor += len + 1;
			}
					
			DragFinish((HDROP) wParam);

			string check = szTarget;
			if (check.substr(1, check.length()-1) == "\\")
				check.append( "*" );
			else
				check.append( "\\*" );
				
			//Test Path for Existence
			WIN32_FIND_DATA FindData;
			HANDLE hFindHandle = FindFirstFile(check.c_str(), &FindData);

			if(hFindHandle != INVALID_HANDLE_VALUE || intFileAction == 0)
			{
				szSourceFiles[scursor] = 0;
				strcpy(buffer, szTarget);
				buffer[strlen(buffer)+1] = 0;
					
				SHFILEOPSTRUCT shfop;
				shfop.hwnd = NULL;
				if (intFileAction == 1)
					shfop.wFunc = FO_MOVE;
				else if (intFileAction == 2)
					shfop.wFunc = FO_COPY;
				else if (intFileAction == 0)
					shfop.wFunc = FO_DELETE;
				shfop.pFrom = szSourceFiles;
				shfop.pTo = buffer;
				shfop.fFlags = FOF_NOCONFIRMMKDIR | FOF_ALLOWUNDO;
				SHFileOperation(&shfop);
			}
				
			delete[] szSourceFiles;
			
			return 0;
		}
		
		case WM_LBUTTONDBLCLK:
		{
			onLButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_LBUTTONDOWN:
		{
			onLButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_LBUTTONUP:
		{
			onLButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONDBLCLK:
		{
			onMButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONDOWN:
		{
			onMButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONUP:
		{
			onMButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONDBLCLK:
		{
			onRButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONDOWN:
		{
			onRButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONUP:
		{
			onRButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MOUSEMOVE:
		{
			if ( TooltipHints )
			{
				MSG hintMessage = {hWnd, message, wParam, lParam};
				SendMessage( TooltipHints, (UINT) TTM_RELAYEVENT, 0, (LPARAM) (LPMSG) &hintMessage ); 
			}

			onMouseMove((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}
		
		case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hDC;

			if(!wParam)
				hDC = BeginPaint(hWnd, &ps);
			else
				hDC = (HDC) wParam;

			onPaint(hDC);

			if(!wParam)
				EndPaint(hWnd, &ps);

			return true;
		}

		case WM_WINDOWPOSCHANGING:
		{
			onWindowPosChanging( (WINDOWPOS *) lParam );
			return false;
		}

		case WM_MOVE:
		{
			onMove( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
			return true;
		}

		case WM_SIZE:
		{
			onSize((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_TIMER:
		{
			onTimer((int) wParam);
			return true;
		}
	}

	return false;
}

LRESULT DesktopItem::windowProcedure(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	DesktopItem *desktopitem = NULL;

	if(message == WM_NCCREATE)
	{
		desktopitem = (DesktopItem *) ((CREATESTRUCT *) lParam)->lpCreateParams;
		desktopitem->hWnd = hWnd;
		SetWindowLong(hWnd, 0, (LONG) desktopitem);
	}
	else
	{
		desktopitem = (DesktopItem *) GetWindowLong(hWnd, 0);
	}

	if(desktopitem)
	{
		LRESULT lResult = 0;

		if(desktopitem->onWindowMessage(message, wParam, lParam, lResult))
			return lResult;
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

//------------------------------------------------------------
//Tooltips (Add, Remove)
//------------------------------------------------------------

void DesktopItem::addHint(HWND hWnd, LPSTR caption)
{
	if (!TooltipHints)
		return;
	
	RECT clientRect;
	GetClientRect(hWnd, &clientRect);

	TOOLINFO ti;
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_TRANSPARENT;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = clientRect;
	ti.hinst = hInstance;
	ti.lpszText = caption;
	ti.lParam = 0;
	SendMessage(TooltipHints, TTM_ADDTOOL, 0, LPARAM(&ti));
}

void DesktopItem::removeHint(HWND hWnd)
{
	if (!TooltipHints)
		return;
	
	TOOLINFO ti;
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = 0;
	ti.hwnd = hWnd;
	ti.uId = 0;
	SetRectEmpty(&ti.rect);
	ti.hinst = hInstance;
	ti.lpszText = NULL;
	ti.lParam = 0;
	SendMessage(TooltipHints, TTM_DELTOOL, 0, LPARAM(&ti));
}

int DesktopItem::isCloseTo(POINT &pos, const int x, const int y, const int span)
{
	bool internalX = false;
	bool internalY = false;

	if ((pos.x > x-span) && (pos.x < x+span))
	{
		if ( (pos.y > y-span) && (pos.y < y+span) )
		{
			pos.x = x;
			internalX = true;
		}
	}

	if ((pos.y > y-span) && (pos.y < y+span))
	{
		if ( (pos.x > x-span) && (pos.x < x+span) )
		{
			pos.y = y;
			internalY = true;
		}
	}

	if (!internalX && !internalY)
		return 0;
	else if (internalX && !internalY)
		return 1;
	else if (!internalX && internalY)
		return 2;
	else
		return 3;
}