/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2005 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "Font.h"

Font::Font()
{
	hFont = 0;
	activehFont = 0;
}

Font::~Font()
{
	if(hFont)
		DeleteObject(hFont);
	if(activehFont)
		DeleteObject(activehFont);
}

void Font::configure(const string &prefix)
{
	shadow = GetxRCBool(prefix, "FontShadow", false);
	shadowColor = GetxRCColor(prefix, "FontShadowColor", RGB(128, 128, 128));
	shadowX = GetxRCInt(prefix, "FontShadowX", 1);
	shadowY = GetxRCInt(prefix, "FontShadowY", 1);

	textoutline = GetxRCBool(prefix, "FontOutline", false);
	textoutlineColor = GetxRCColor(prefix, "FontOutlineColor", RGB(128, 128, 128));

	fontname = GetxRCString(prefix, "Font", "Arial");
	fontheight = GetxRCInt(prefix, "FontHeight", 15, 6);
	fontcolor = GetxRCColor(prefix, "FontColor", RGB(0, 0, 0));

	bold = GetxRCBool(prefix, "FontBold", false);
	italic = GetxRCBool(prefix, "FontItalic", false);
	underline = GetxRCBool(prefix, "FontUnderline", false);

	activeshadow = GetxRCBool(prefix, "ActiveFontShadow", shadow);
	activeshadowColor = GetxRCColor(prefix, "ActiveFontShadowColor", shadowColor);
	activeshadowX = GetxRCInt(prefix, "ActiveFontShadowX", shadowX);
	activeshadowY = GetxRCInt(prefix, "ActiveFontShadowY", shadowY);

	activetextoutline = GetxRCBool(prefix, "ActiveFontOutline", textoutline);
	activetextoutlineColor = GetxRCColor(prefix, "ActiveFontOutlineColor", textoutlineColor);

	activefontname = GetxRCString(prefix, "ActiveFont", fontname);
	activefontheight = GetxRCInt(prefix, "ActiveFontHeight", fontheight, 6);
	activefontcolor = GetxRCColor(prefix, "ActiveFontColor", fontcolor);

	activebold = GetxRCBool(prefix, "ActiveFontBold", bold);
	activeitalic = GetxRCBool(prefix, "ActiveFontItalic", italic);
	activeunderline = GetxRCBool(prefix, "ActiveFontUnderline", underline);

	createHandle(true);
	createHandle(false);
}

void Font::apply(HDC hDC, int x, int y, int width, int height, const string &text, unsigned int flags, bool active)
{
	HFONT oldhFont = (HFONT) SelectObject(hDC, (active ? activehFont : hFont));
	int oldbkMode = SetBkMode(hDC, TRANSPARENT);
	int oldtextColor = SetTextColor(hDC, (active ? activefontcolor : fontcolor));
		
	RECT r;
	r.left = x;	
	r.right = r.left + width;
	r.top = y;
	r.bottom = r.top + height;

	if ( (active && activeshadow && !activetextoutline ) || (!active && shadow && !textoutline) )
	{
		OffsetRect(&r, ((active ? activeshadowX : shadowX) + 1) / 2, ((active ? activeshadowY : shadowY)+ 1) / 2);
		SetTextColor(hDC, (active ? activeshadowColor : shadowColor));
		DrawTextEx(hDC, (char *) text.c_str(), text.length(), &r, flags | DT_END_ELLIPSIS, 0);
		OffsetRect(&r, -(active ? activeshadowX : shadowX), -(active ? activeshadowY : shadowY));
	}
	if ( (active && activetextoutline) || ( !active && textoutline) )
	{
		RECT tmpr;
		CopyRect(&tmpr, &r);
		SetTextColor(hDC, (active ? activetextoutlineColor : textoutlineColor));

		OffsetRect(&tmpr, 1, 0);
		DrawTextEx(hDC, (char *) text.c_str(), text.length(), &tmpr, flags | DT_END_ELLIPSIS, 0);

		OffsetRect(&tmpr, -2, 0);
		DrawTextEx(hDC, (char *) text.c_str(), text.length(), &tmpr, flags | DT_END_ELLIPSIS, 0);

		OffsetRect(&tmpr, 1, -1);
		DrawTextEx(hDC, (char *) text.c_str(), text.length(), &tmpr, flags | DT_END_ELLIPSIS, 0);

		OffsetRect(&tmpr, 0, 2);
		DrawTextEx(hDC, (char *) text.c_str(), text.length(), &tmpr, flags | DT_END_ELLIPSIS, 0);
	}

	SetTextColor(hDC, (active ? activefontcolor : fontcolor));

	DrawTextEx(hDC, (char *) text.c_str(), text.length(), &r, flags | DT_END_ELLIPSIS, 0);

	if ( active )
		activehFont = (HFONT) SelectObject(hDC, oldhFont);
	else
		hFont = (HFONT) SelectObject(hDC, oldhFont);

	SetBkMode(hDC, oldbkMode);
	SetTextColor(hDC, oldtextColor);
}

POINT Font::measure(HDC hDC, int x, int y, int width, int height, const string &text, unsigned int flags, bool active)
{
	POINT maxValues = {0,0};
	RECT r;
	r.left = x;
	r.right = r.left + width;
	r.top = y;
	r.bottom = r.top + height;

	if (active)
	{
		HFONT oldhFont = (HFONT) SelectObject(hDC, activehFont);

		DrawTextEx(hDC, (char *) text.c_str(), text.length(), &r, flags | DT_CALCRECT, 0);
	
		maxValues.x = (r.right-r.left);
		maxValues.y = (r.bottom-r.top);

		activehFont = (HFONT) SelectObject(hDC, oldhFont);
	}
	else 
	{
		HFONT oldhFont = (HFONT) SelectObject(hDC, hFont);

		DrawTextEx(hDC, (char *) text.c_str(), text.length(), &r, flags | DT_CALCRECT, 0);
	
		maxValues.x = (r.right-r.left);
		maxValues.y = (r.bottom-r.top);

		hFont = (HFONT) SelectObject(hDC, oldhFont);
	}

	return maxValues;
}

void Font::createHandle(bool active)
{
	LOGFONT lf;
	memset(&lf, 0, sizeof(LOGFONT));

	if ( active )
	{
		StringCchCopy(lf.lfFaceName, LF_FACESIZE, activefontname.c_str());
		lf.lfHeight = activefontheight;
		lf.lfItalic = activeitalic ? TRUE : FALSE;
		lf.lfUnderline = activeunderline ? TRUE : FALSE;
		lf.lfWeight = activebold ? FW_BOLD : FW_NORMAL;
		lf.lfCharSet = DEFAULT_CHARSET;
		lf.lfQuality = NONANTIALIASED_QUALITY;

		if(activehFont)
			DeleteObject(activehFont);

		activehFont = CreateFontIndirect(&lf);
	}
	else
	{
		StringCchCopy(lf.lfFaceName, LF_FACESIZE, fontname.c_str());
		lf.lfHeight = fontheight;
		lf.lfItalic = italic ? TRUE : FALSE;
		lf.lfUnderline = underline ? TRUE : FALSE;
		lf.lfWeight = bold ? FW_BOLD : FW_NORMAL;
		lf.lfCharSet = DEFAULT_CHARSET;
		lf.lfQuality = NONANTIALIASED_QUALITY;

		if(hFont)
			DeleteObject(hFont);

		hFont = CreateFontIndirect(&lf);
	}
}
