/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2005 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "StdAfx.h"
#include "DesktopItem.h"
#include "DesktopTexture.h"
#include "Font.h"

#include "FileInformation.h"
#include "NotifyDirCheck.h"

#include "regexp.h"
#include "regmagic.h"

void getToken(char line[MAX_LINE_LENGTH], char temp[MAX_LINE_LENGTH], bool escape_codes = true);

void TextAppend(HWND hwndCaller, LPCSTR pszArgs);
void TextDelete(HWND hwndCaller, LPCSTR pszArgs);

StringList scanDesktopItemPaths(string searchPath, bool defaultDesktop = false);
void internalDesktopUpdate(int updateMode, string updatedFile, StringList newDesktopItemPaths);

void BangDestroy(HWND hwndCaller, LPCSTR pszArgs);
void BangCreate(HWND hwndCaller, LPCSTR pszArgs);
void BangHide(HWND hwndCaller, LPCSTR pszArgs);
void BangShow(HWND hwndCaller, LPCSTR pszArgs);
void BangToggle(HWND hwndCaller, LPCSTR pszArgs);

void OnStart(LPCSTR path);
void OnStop();

bool useautodesktop = GetxRCBool("xDesktop", "AutoCreate");
string desktopSavePath = GetxRCString("xDesktop", "ConfigFile", "");
bool disableSystemContext = GetxRCBool("xDesktop", "NoSystemContextMenu");

class CNotifyDirCheck;

typedef list<CNotifyDirCheck *> DirCheckList;
typedef list<CNotifyDirCheck *>::iterator DirCheckListIterator;

DirCheckList dircheckList;
CStringList pathList;

DesktopItemList desktopitemList;

Font *defaultfont = NULL;
DesktopTexture *defaultbackground = NULL;

HINSTANCE hInstance;
HWND messageHandler;

int lsMessages[] = {
	LM_GETREVID,
	0
};

#define SM_CXVIRTUALSCREEN      78
#define SM_CYVIRTUALSCREEN      79

int screenX = GetSystemMetrics(SM_CXVIRTUALSCREEN);
int	screenY = GetSystemMetrics(SM_CYVIRTUALSCREEN);

int bitsperpixel = atoi( QueryPixelDepth().c_str() );

#define LM_UPDATEBG (WM_USER + 1)

LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case LM_GETREVID:
		{
			UINT uLength;
			StringCchPrintf((char*)lParam, 64, "%s.dll: %s", V_NAME, V_VERSION);
			
			if (SUCCEEDED(StringCchLength((char*)lParam, 64, &uLength)))
				return uLength;

			lParam = NULL;
			return 0;
		}

		case LM_UPDATEBG:
		{
			PaintDesktopEx(0, 0, 0, 0, 0, 0, 0, TRUE);

			for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
				(*diit)->repaint(true, true);

			return 0;
		}

		case WM_SETTINGCHANGE:
		{
			PostMessage(hWnd, LM_UPDATEBG, 0, 0);
			return 0;
		}
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

// Why are we advoiding the crt startup code? The start up code is not large
// enough to cause bloat (except possibly a long system call, and I suspect that
// would need done eventually anyway) (default _DllMainCRTStartup is in
// (MSVCDIR\crt\src\dllcrt0.c) The elimination of the call to _CRT_INIT is 
// also causing warnings with MSVC7
BOOL __stdcall _DllMainCRTStartup(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes)
{
	// We don't need thread notifications for what we're doing.  Thus, get
	// rid of them, thereby eliminating some of the overhead of this DLL
	DisableThreadLibraryCalls(hInst);
	return TRUE;
}

int initModuleEx(HWND hParent, HINSTANCE hInstance, const char *lsPath)
{
	WNDCLASSEX wc;

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS | CS_DBLCLKS;
	wc.lpfnWndProc = DesktopItem::windowProcedure;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = sizeof(DesktopItem *);
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = 0;
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xDesktopItem";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS;
	wc.lpfnWndProc = MessageHandlerProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = 0;
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xDesktopItemMessageHandler";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	messageHandler = CreateWindowEx(WS_EX_TOOLWINDOW,
		"xDesktopItemMessageHandler",
		0,
		WS_POPUP,
		0, 0, 0, 0, 
		0,
		0,
		hInstance,
		0);

	if (!messageHandler)
		return 1;
	
	SendMessage(GetLitestepWnd(),
		LM_REGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	::hInstance = hInstance;

	if ( desktopSavePath.empty() )
	{
		MessageBox(NULL, "You haven't defined the setting xDesktopConfigFile, which points to the file containing your *xDesktopItem lines!\nxDesktop will now Quit!!", "Important User Information", MB_SETFOREGROUND | MB_ICONSTOP );

		SendMessage(GetLitestepWnd(),
			LM_UNREGISTERMESSAGE,
			(WPARAM) messageHandler,
			(LPARAM) lsMessages);

		DestroyWindow(messageHandler);

		UnregisterClass("xDesktopItem", hInstance);
		UnregisterClass("xDesktopItemMessageHandler", hInstance);

		return 1;
	}

	//Default Font
	defaultfont = new Font();
	defaultfont->configure("xDesktop");
	//Default Texture
	defaultbackground = new DesktopTexture();
	defaultbackground->configure();

	//Desktop Items
	StringList stringDesktopItemList = GetxRCLineList("*xdesktopitem");

	if (!stringDesktopItemList.empty())
	{	
		for(StringListIterator sit = stringDesktopItemList.begin(); sit != stringDesktopItemList.end(); sit++)
		{
			char name[32];
			char *buffers[] = {name};

			if(LCTokenize((*sit).c_str(), buffers, 1, 0) >= 1)
			{
				DesktopItem *desktopitem = new DesktopItem(name);

				desktopitemList.insert(desktopitemList.end(), desktopitem);

				desktopitem->desktopload(hInstance, (*sit));
			}
		}
	}

	AddBangCommand("!xDesktopCreate", BangCreate);
	AddBangCommand("!xDesktopDestroy", BangDestroy);
	AddBangCommand("!xDesktopHide", BangHide);
	AddBangCommand("!xDesktopShow", BangShow);
	AddBangCommand("!xDesktopToggle", BangToggle);

	if (useautodesktop)
	{
		if (stringDesktopItemList.empty())
			LSExecute(NULL, ("!xdesktopcreate replace \".desktop\""), SW_SHOWNORMAL);

		char path[MAX_PATH] = {0};			
		
		if ( IsOS( OS_2KXP ) )
		{
			VarExpansion( path, "$commondesktopdir$" );
			OnStart( path );
		}

		VarExpansion( path, "$desktopdir$" );
		OnStart( path );
	}

	return 0;
}

extern HDC hdcDesktop;
extern HBITMAP hbmDesktop;

void quitModule(HINSTANCE hInstance)
{
	RemoveBangCommand("!xDesktopCreate");
	RemoveBangCommand("!xDesktopDestroy");
	RemoveBangCommand("!xDesktopHide");
	RemoveBangCommand("!xDesktopShow");
	RemoveBangCommand("!xDesktopToggle");

	//Clear all *xdesktopitem lines
	TextDelete(NULL, ("@" + desktopSavePath + "@ @\\*[xX][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]@").c_str());

	int counter = 1;
	char buffer[32] = {0};	
	char counterbuffer[16] = {0};

	for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
	{
		strcpy(buffer, "xdi");
		itoa( counter, counterbuffer, 10 );
		if (counter < 10)
			strcat(buffer, "0");
		strcat(buffer, counterbuffer);

		TextAppend(NULL, ("@" + desktopSavePath + "@ @*xdesktopitem " + buffer + " " + (*diit)->getPositionSaver() + "@").c_str());

		counter++;

		delete *diit;
	}

	desktopitemList.clear();

	if ( defaultfont )
		delete defaultfont;

	if ( defaultbackground )
		delete defaultbackground;

	if (useautodesktop)
		OnStop();

	SendMessage(GetLitestepWnd(),
		LM_UNREGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	DestroyWindow(messageHandler);

	UnregisterClass("xDesktopItem", hInstance);
	UnregisterClass("xDesktopItemMessageHandler", hInstance);

	if (hbmDesktop != 0 || hdcDesktop != 0)
	{
		hbmDesktop = (HBITMAP) SelectObject(hdcDesktop, hbmDesktop);
		DeleteDC(hdcDesktop);
		DeleteObject(hbmDesktop);
	}
}

void RemoveDesktopItem(string path)
{
	for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
	{
		if (stricmp((*diit)->getIconCommand().c_str(), path.c_str()) == NULL)
		{
			delete (*diit);
			desktopitemList.remove( *diit );
			
			break;
		}
	}
}

UINT DirCallback( CFileInformation fiObject, EFileAction faAction, LPVOID lpData )
{
	CString filepath = fiObject.GetFilePath();

	bool found = false;
	POSITION m_position = pathList.GetHeadPosition();
	while (m_position != NULL && found == FALSE)
	{
		if ( fiObject.IsSomeFileDir( pathList.GetAt(m_position) ) )
			found = TRUE;
		pathList.GetNext(m_position);
	}

	if (found)
	{
		if( IS_CREATE_FILE( faAction ) )
		{
			ParseBangCommand(NULL, "!xdesktopcreate", ("updateadd \"" + filepath + "\""));
		}
		else if( IS_DELETE_FILE( faAction ) )
		{
			ParseBangCommand(NULL, "!xdesktopcreate", ("updatedel \"" + filepath + "\""));
		}
		else if( IS_CHANGE_FILE( faAction ) )
		{
			ParseBangCommand(NULL, "!xdesktopcreate", ("updateren \"" + filepath + "\""));
		}
		else
		{
			return 1; //error, stop thread
		}
	}

	return 0; //success
}

void OnStart(LPCSTR path) 
{
	CNotifyDirCheck *DirCheck = new CNotifyDirCheck();
	
	dircheckList.insert(dircheckList.end(), DirCheck);
		
	DirCheck->SetDirectory( path );
	pathList.AddTail( path );

	//set your callback to work with each new event
	DirCheck->SetActionCallback( DirCallback );

	DirCheck->Run();
}

void OnStop() 
{
	for(DirCheckListIterator dcit = dircheckList.begin(); dcit != dircheckList.end(); dcit++)
	{
		(*dcit)->Stop();
		delete *dcit;
	}

	pathList.RemoveAll();
	dircheckList.clear();
}

//Read all required Files/Folder Paths (in a StringList)
StringList scanDesktopItemPaths(string searchPath, bool defaultDesktop)
{
	StringList returnDesktopItemPaths;

	string itemName = "";
	
	char tmp[MAX_PATH] = {0};
	string tmpHelper = "";

	HANDLE hFindHandle;
	WIN32_FIND_DATA FindData;

	if (defaultDesktop)
	{
		//Create 4 Main ShellFolders: Workspace, My Documents, NetworkHood, Trash 
		if ( !GetxRCBool("xDesktop", "NoWorkSpace") )
			returnDesktopItemPaths.insert(returnDesktopItemPaths.end(), "::{20D04FE0-3AEA-1069-A2D8-08002B30309D}");
		if ( !GetxRCBool("xDesktop", "NoMyDocuments") )
			returnDesktopItemPaths.insert(returnDesktopItemPaths.end(), "::{450D8FBA-AD25-11D0-98A8-0800361B1103}");
		if ( !GetxRCBool("xDesktop", "NoNetwork") )
			returnDesktopItemPaths.insert(returnDesktopItemPaths.end(), "::{208D2C60-3AEA-1069-A2D7-08002B30309D}");
		if ( !GetxRCBool("xDesktop", "NoTrashcan") )
			returnDesktopItemPaths.insert(returnDesktopItemPaths.end(), "::{645FF040-5081-101B-9F08-00AA002F954E}");

		if ( IsOS( OS_2KXP ) )
		{
			VarExpansion( tmp, "$commondesktopdir$" );
			searchPath = tmp;
			tmpHelper = (searchPath + "*.*");

			hFindHandle = FindFirstFile(tmpHelper.c_str(), &FindData);

			for(;;)
			{
				if(INVALID_HANDLE_VALUE == hFindHandle)
					break;

				itemName = FindData.cFileName;

				//Make sure its a valid folder
				if (stricmp(itemName.c_str(), ".") != NULL && stricmp(itemName.c_str(), "..") != NULL)
				{
					tmpHelper = (searchPath + itemName);
					returnDesktopItemPaths.insert(returnDesktopItemPaths.end(), tmpHelper.c_str());
				}

				if(!FindNextFile(hFindHandle, &FindData))
					break;
			}
		}

		VarExpansion( tmp, "$desktopdir$" );
		searchPath = tmp;
		tmpHelper = (searchPath + "*.*");

		hFindHandle = FindFirstFile(tmpHelper.c_str(), &FindData);

		for(;;)
		{
			if(INVALID_HANDLE_VALUE == hFindHandle)
				break;
					
			itemName = FindData.cFileName;

			//Make sure its a valid folder
			if (stricmp(itemName.c_str(), ".") != NULL && stricmp(itemName.c_str(), "..") != NULL)
			{
				tmpHelper = (searchPath + itemName);
				returnDesktopItemPaths.insert(returnDesktopItemPaths.end(), tmpHelper.c_str());
			}

			if(!FindNextFile(hFindHandle, &FindData))
				break;
		}
	}
	else
	{
		//Check for ending BackSlash
		tmpHelper = searchPath.substr(searchPath.length()-1, searchPath.length());

		if (tmpHelper != "\\")
			tmpHelper = (searchPath + "\\*.*");
		else
			tmpHelper = (searchPath + "*.*");

		hFindHandle = FindFirstFile(tmpHelper.c_str(), &FindData);

		for(;;)
		{
			if(INVALID_HANDLE_VALUE == hFindHandle)
				break;
					
			itemName = FindData.cFileName;

			//Make sure its a valid folder
			if (stricmp(itemName.c_str(), ".") != NULL && stricmp(itemName.c_str(), "..") != NULL)
			{
				tmpHelper = (searchPath + itemName);
				returnDesktopItemPaths.insert(returnDesktopItemPaths.end(), tmpHelper.c_str());
			}

			if(!FindNextFile(hFindHandle, &FindData))
				break;
		}
	}

	FindClose(hFindHandle);

	return returnDesktopItemPaths;
}

//Write all Items to File
void internalDesktopUpdate(int updateMode, string updatedFile, StringList newDesktopItemPaths)
{
	int counter = 1;

	//Saves already present Items, Only if:
	//	1. They exist
	//  2. It's wished
	//  3. AutoCreateDesktop isn't running
	if ( !desktopitemList.empty() && updateMode == 1 && !useautodesktop )
	{
		char buffer[32] = {0};
		char counterbuffer[16] = {0};
		for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
		{
			strcpy(buffer, "xdi");
			itoa( counter, counterbuffer, 10 );
			if (counter < 10)
				strcat(buffer, "0");
			strcat(buffer, counterbuffer);

			TextAppend(NULL, ("@" + desktopSavePath + "@ @*xdesktopitem " + buffer + " " + (*diit)->getPositionSaver() + "@").c_str());

			counter++;
		}
	}

	//Saves new Items, Only if:
	//	1. It's wished (Replace)
	//  2. It's wished (Append)
	//  3. None exists (Initial Creation required)
	//  AND IT'S NOT AN UPDATE
	if ( (updateMode == 0 || updateMode == 1 || desktopitemList.empty()) && updateMode != -1)
	{
		int tempcounter = 1;

		NameValuePair directionValues[] = {
			{ "left", 1 },
			{ "right", 2 },
			{ "up", 3 },
			{ "down", 4 },
			{ 0, 0 }
			};

		NameValuePair wrapdirectionValues[] = {
			{ "up", 1 },
			{ "down", 2 },
			{ "left", 3 },
			{ "right", 4 },
			{ 0, 0 }
			};

		int DIwidth = GetxRCInt("xDesktop", "ItemWidth", 100);
		int DIheight = GetxRCInt("xDesktop", "ItemHeight", 75);
		int xSpace = GetxRCInt("xDesktop", "SpacingX", 0);
		int ySpace = GetxRCInt("xDesktop", "SpacingY", 0);
				
		int direction = GetxRCNamedValue("xDesktop", "Direction", directionValues, 2);
		int wrapDirection = GetxRCNamedValue("xDesktop", "WrapDirection", wrapdirectionValues, 2);
		
		int wrapCount;
		if (wrapDirection == 1 || wrapDirection == 2)
			wrapCount = GetxRCInt("xDesktop", "WrapCount", (screenX-xSpace)/(DIwidth+xSpace));
		else if (wrapDirection == 3 || wrapDirection == 4)
			wrapCount = GetxRCInt("xDesktop", "WrapCount", (screenY-ySpace)/(DIheight+ySpace));

		int DIStartx = GetxRCCoord("xDesktop", "X", ((screenX-(wrapCount*(DIwidth+xSpace)-xSpace))/2 < 0 ? 0 : (screenX-(wrapCount*(DIwidth+xSpace)-xSpace))/2), screenX);
		int DIStarty = GetxRCCoord("xDesktop", "Y", ((screenY-(wrapCount*(DIheight+ySpace)-ySpace))/2 < 50 ? 50 : (screenY-(wrapCount*(DIheight+ySpace)-ySpace))/2), screenY);

		int tempx;
		int tempy;

		if (direction == 1)
			tempx = DIStartx+DIwidth+xSpace;
		else if (direction == 2)
			tempx = DIStartx-(DIwidth+xSpace);
		else if (direction == 3)
			tempy = DIStarty+DIheight+ySpace;
		else if (direction == 4)
			tempy = DIStarty-(DIheight+ySpace);

		if (direction == 1 || direction == 2)
			tempy = DIStarty;
		else if (direction == 3 || direction == 4)
			tempx = DIStartx;

		char buffer[32] = {0};
		char counterbuffer[16] = {0};
		char xbuffer[16] = {0};
		char ybuffer[16] = {0};
		for(StringListIterator sit = newDesktopItemPaths.begin(); sit != newDesktopItemPaths.end(); sit++)
		{
			strcpy(buffer, "xdi");
			itoa( counter, counterbuffer, 10 );
			if (counter < 10)
				strcat(buffer, "0");
			strcat(buffer, counterbuffer);

			//Sorting/Positioning due to settings
			div_t div_result = div( tempcounter, wrapCount );

			if (direction == 1 || direction == 2)
			{
				if (wrapDirection > 2)
					wrapDirection = 2;
						
				if (div_result.quot > 0 && div_result.rem == 1)
				{
					tempx = DIStartx;
					if (wrapDirection == 1)
						tempy = DIStarty-(div_result.quot*(DIheight+ySpace));
					else if (wrapDirection == 2)
						tempy = DIStarty+div_result.quot*(DIheight+ySpace);
				}
				else
				{
					if (direction == 1)
						tempx = tempx-(DIwidth+xSpace);
					else if (direction == 2)
						tempx = tempx+DIwidth+xSpace;
				}
			}
			if (direction == 3 || direction == 4)
			{
				if (wrapDirection < 3)
					wrapDirection = 4;
					
				if (div_result.quot > 0 && div_result.rem == 1)
				{
					tempy = DIStarty;
					if (wrapDirection == 3)
						tempx = DIStartx-(div_result.quot*(DIwidth+xSpace));
					else if (wrapDirection == 4)
						tempx = DIStartx+div_result.quot*(DIwidth+xSpace);
				}
				else
				{
					if (direction == 3)
						tempy = tempy-(DIheight+ySpace);
					else if (direction == 4)
						tempy = tempy+DIheight+ySpace;
				}
			}

			itoa( tempx, xbuffer, 10 );
			itoa( tempy, ybuffer, 10 );

			if (!useautodesktop)
				TextAppend(NULL, ("@" + desktopSavePath + "@ @*xdesktopitem " + buffer + " " + xbuffer + " " + ybuffer + " \"" + (*sit) + "\"@").c_str());

			DesktopItem *desktopitem = new DesktopItem(buffer);

			desktopitemList.insert(desktopitemList.end(), desktopitem);

			string Bang = buffer;
			Bang.append( " " );
			Bang.append( xbuffer );
			Bang.append( " ");
			Bang.append( ybuffer );
			Bang.append( " \"" + (*sit) + "\"");

			desktopitem->desktopload(hInstance, Bang.c_str());

			counter++;
			tempcounter++;
		}
	}
	//Update & Saves new Items, Only if:
	//	1. Some have changed
	else if (updateMode == -1)
	{
		StringList updatedDesktopItemPaths;

		for (StringListIterator sit = newDesktopItemPaths.begin(); sit != newDesktopItemPaths.end(); sit++)
		{
			bool found = false;

			for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
			{
				if (stricmp((*sit).c_str(), (*diit)->getIconCommand().c_str()) == NULL)
				{
					found = true;
					break;
				}
			}

			if (found)
			{
				updatedDesktopItemPaths.insert(updatedDesktopItemPaths.end(), ((*diit)->getPositionSaver()));

				delete *diit;
				desktopitemList.remove( *diit );					
			}
			else
			{
				updatedDesktopItemPaths.insert(updatedDesktopItemPaths.end(), ("100 100 \"" + (*sit) + "\""));
			}
		}

		for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
			delete *diit;

		desktopitemList.clear();
		
		int counter = 1;
		char buffer[32] = {0};
		char counterbuffer[16] = {0};
		for(sit = updatedDesktopItemPaths.begin(); sit != updatedDesktopItemPaths.end(); sit++)
		{
			strcpy(buffer, "xdi");
			itoa( counter, counterbuffer, 10 );
			if (counter < 10)
				strcat(buffer, "0");
			strcat(buffer, counterbuffer);

			DesktopItem *desktopitem = new DesktopItem(buffer);

			desktopitemList.insert(desktopitemList.end(), desktopitem);

			desktopitem->desktopload(hInstance, ("dummy " + (*sit)));
			
			counter++;
		}
	}
	//Update & Saves single changed Items, Only if:
	//	1. they have changed
	//	1. AutoCreateDesktop is ON
	else if (updateMode >= 20 && useautodesktop)
	{
		//Add Files
		if (updateMode == 21)
		{
			string Bang = ( "dummy 100 100 \"" + updatedFile + "\"" );

			DesktopItem *desktopitem = new DesktopItem("dummy");

			desktopitemList.insert(desktopitemList.end(), desktopitem);

			desktopitem->desktopload(hInstance, Bang.c_str());
		}
		//Remove Files
		else if (updateMode == 22)
		{
			for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
			{
				if (stricmp((*diit)->getIconCommand().c_str(), updatedFile.c_str()) == NULL)
				{
					delete *diit;
					desktopitemList.remove( *diit );
					
					break;
				}
			}
		}
		//Rename Files
		else if (updateMode == 23)
		{
			HANDLE hFindHandle;
			WIN32_FIND_DATA FindData;

			for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
			{
				if (strstr((*diit)->getIconCommand().c_str(), "::{") == NULL)
				{
					hFindHandle = FindFirstFile((*diit)->getIconCommand().c_str(), &FindData);

					if(INVALID_HANDLE_VALUE == hFindHandle)
					{
						string name = (*diit)->getName();
						string positions = ( IntToString((*diit)->getX()) + " " + IntToString((*diit)->getY()) );
						
						delete *diit;
						desktopitemList.remove( *diit );
						
						string Bang = ( name + " " + positions + " \"" + updatedFile + "\"" );

						DesktopItem *desktopitem = new DesktopItem(name);

						desktopitemList.insert(desktopitemList.end(), desktopitem);

						desktopitem->desktopload(hInstance, Bang.c_str());

						break;
					}
				}
			}

			FindClose(hFindHandle);
		}
	}
}

// !xDesktopCreate <append>|<replace>|<update...> <path to lnkfolder>|<updatefile>
void BangCreate(HWND hwndCaller, LPCSTR pszArgs)
{
	StringList newDesktopItemPaths;

	char action[32], path[MAX_LINE_LENGTH];
	char *buffers[] = {action, path};

	LCTokenize(pszArgs, buffers, 2, NULL);
	
	string searchPath = path;

	//Quit here, if no path is specified
	if (searchPath.empty())
		return;

	//Initialize
	bool defaultDesktop = false;
	
	if ( stricmp(searchPath.c_str(), ".desktop") == 0 || stricmp(searchPath.c_str(), "deskto") == 0 || stricmp(searchPath.c_str(), "eskto") == 0 )
		defaultDesktop = true;
	
	//Start Configuration and Writing to File
	int updateMode = 0;

	if ( stricmp(action, "update") == NULL )
		updateMode = -1;
	else if ( stricmp(action, "updateadd") == NULL )
		updateMode = 21;
	else if ( stricmp(action, "updatedel") == NULL )
		updateMode = 22;
	else if ( stricmp(action, "updateren") == NULL )
		updateMode = 23;
	else if (stricmp(action, "append") == NULL)
		updateMode = 1;
	else //Replace
		updateMode = 0;

	if (updateMode == -1 || updateMode == 0 || updateMode == 1)
	{
		//Get All Items actual items
		newDesktopItemPaths = scanDesktopItemPaths(searchPath, defaultDesktop);

		if (updateMode == 0)
			LSExecute(NULL, "!xdesktopdestroy purge", SW_SHOWNORMAL);

		//Clear all *xdesktopitem lines
		TextDelete(NULL, ("@" + desktopSavePath + "@ @\\*[xX][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]@").c_str());
	}

	//0 = Create New (Replace/Reposition), 1 = Append (New will be Positioned), 2+ = Update
	internalDesktopUpdate(updateMode, searchPath, newDesktopItemPaths);
}

// !xDesktopDestroy
void BangDestroy(HWND hwndCaller, LPCSTR pszArgs)
{	
	bool purge = false;
	if ( stricmp(pszArgs, "purge") != NULL )
		purge = true;

	//Clear all *xdesktopitem lines
	if ( !purge )
		TextDelete(NULL, ("@" + desktopSavePath + "@ @\\*[xX][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]@").c_str());

	int counter = 1;
	char buffer[32] = {0};	
	char counterbuffer[16] = {0};
	for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
	{
		if ( !purge )
		{
			strcpy(buffer, "xdi");
			itoa( counter, counterbuffer, 10 );
			if (counter < 10)
				strcat(buffer, "0");
			strcat(buffer, counterbuffer);

			TextAppend(NULL, ("@" + desktopSavePath + "@ @*xdesktopitem " + buffer + " " + (*diit)->getPositionSaver() + "@").c_str());
		
			counter++;
		}

		delete *diit;
	}

	desktopitemList.clear();
}

// !xDesktopHide
void BangHide(HWND hwndCaller, LPCSTR pszArgs)
{
	for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
		(*diit)->hide();
}

// !xDesktopShow
void BangShow(HWND hwndCaller, LPCSTR pszArgs)
{
	for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
		(*diit)->show();
}

// !xDesktopToggle
void BangToggle(HWND hwndCaller, LPCSTR pszArgs)
{
	for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
	{
		if((*diit)->isVisible())
			(*diit)->hide();
		else
			(*diit)->show();
	}
}

void getToken(char line[MAX_LINE_LENGTH], char temp[MAX_LINE_LENGTH], bool escape_codes)
{
    int offset = 0, i;
	bool endquote = false;
    char ENCAP = '@';

	while (line[offset] == ' ')
	{
		offset++;
	}

	if (line[offset] == ENCAP)
	{
		offset++;
		endquote = true;
	}

	for (i = 0; i+offset < int(strlen(line)); ++i)
	{
		// Copy else if block to add more escape codes.
		if (escape_codes && (line[i+offset] == '\\' && int(strlen(line)) != i+offset+1))
		{
		    if (line[i+offset+1] == ENCAP || line[i+offset+1] == '\\')
			{
				offset++;
				temp[i] = line[i+offset];
				continue;
			}
		}

 	    if (line[i+offset] == ENCAP && endquote)
	  	    break;
		else if (line[i+offset] == ' ' && !endquote)
			break;
		else
			temp[i] = line[i+offset];
	}
	temp[i] = '\0';

	if (i+offset+1 > int(strlen(line)))
	{
		line[0] = '\0';
	}
	else
	{
		strcpy(line, line+i+offset+1);
	}
}

void TextAppend(HWND hwndCaller, LPCSTR pszArgs)
{
	char filename[MAX_LINE_LENGTH], str[MAX_LINE_LENGTH];
	char rest[MAX_LINE_LENGTH];
	strcpy(rest, pszArgs);

	getToken(rest, filename, false);
	getToken(rest, str, false);

	ofstream file;
	file.open(filename, ios::app);

	if (!file.fail())
	{
		file << str << endl;
	}

	file.close();
}

void TextDelete(HWND hwndCaller, LPCSTR pszArgs)
{
	char filename[MAX_LINE_LENGTH], regex[MAX_LINE_LENGTH], input[MAX_LINE_LENGTH];
	char rest[MAX_LINE_LENGTH];
	regexp *r;
	strcpy(rest, pszArgs);

	getToken(rest, filename, false);
	getToken(rest, regex, true);

	r = regcomp(regex);

	ifstream file;
	file.open(filename);

	if (!file.fail())
	{
		list<string> l;
		string s;
		while (file.getline(input, MAX_LINE_LENGTH))
		{
			if (!regexec(r, input))
			{
				s = input;
				l.push_back(s);
			}
		}
		file.close();
		ofstream outfile;
		outfile.open(filename);
		for (list<string>::const_iterator lci = l.begin(); lci!=l.end(); lci++)
		{
			outfile << lci->data() << endl;
		}
		l.clear();
	}
}

